//Helper functions to deal with file I/O.
//This is a port of the Dojo build's fileUtil from Rhino
//to AIR.

var fileUtil = {};

//fileUtil.baseDir = air.File.documentsDirectory;
fileUtil.baseDir = air.File.userDirectory;
fileUtil.basePath = "airbuild/buildscripts/";

fileUtil.getFileObj = function(/*String or File*/obj){
	if(!obj.nativePath){
		/* Check to see if the base directory is already attached */
		// strip off a file: prefix before comparing
		if(obj.indexOf("file:/") == 0){
			obj = obj.substring(5);
		}
		if(obj.indexOf(fileUtil.baseDir.nativePath) == 0){
			return new air.File(obj);
		}
		
		/* Since we don't have a known "working directory" like the
		   command line tool, we'll assume a path relative to the
 		   	buildscripts path. */
		return fileUtil.baseDir.resolvePath(fileUtil.basePath).resolvePath(obj);
	}
	return obj;
}
fileUtil.fileExists = function(/*String or File*/obj){
	obj = fileUtil.getFileObj(obj);
	return obj.exists;
};

fileUtil.getFileWriter = function(/*String or File*/obj){
	var filename = fileUtil.getFileObj(obj).nativePath;
	logger.trace("making FileWriter for " + filename);
	
	var fs = new function() {
		this.filename = filename;
		this.contents = [];
	};
	fs.write = function(str) {
		this.contents.push(str);
	};
	fs.close = function() {
		fileUtil.saveUtf8File(this.filename, this.contents.join(""));
	};
	return fs;
}

fileUtil.getLineSeparator = function(){
	//summary: Gives the line separator for the platform.
	//could not find this in AIR
	return "\n";
}

fileUtil.getFilteredFileList = function(/*String*/startDir, /*RegExp|String*/regExpFilter, /*boolean?*/makeUnixPaths, /*boolean?*/startDirIsJavaObject){
	//summary: Recurses startDir and finds matches to the files that match regExpFilter.
	//Ignores files/directories that start with a period (.).
	//This function is called recursively with startDir being a
	//two element array. The first element is the original
	//root directory (for computing relative paths).
	//The second is the directory to start searching from.
	
	if(typeof(regExpFilter) == "string"){
		if(regExpFilter.charAt(0) == '/' && regExpFilter.charAt(regExpFilter.length-1) == '/') {
			regExpFilter.substring(1, regExpFilter.length-2);
		}
		regExpFilter = new RegExp(regExpFilter);
	}
	var files = [];
	
	if(startDir instanceof Array){
		var initialRoot = startDir[0];
		var topDir = startDir[1];
	} else {
		var topDir = fileUtil.getFileObj(startDir);
		var initialRoot = topDir;
	}
	

	if( fileUtil.fileExists(topDir) ){
		var dirFileArray = topDir.getDirectoryListing();
		for (var i = 0; i < dirFileArray.length; i++){
			var file = dirFileArray[i];
			if(!file.isDirectory){
				var filePath = file.nativePath;
				if(!file.name.match(/^\./) && filePath.match(regExpFilter)){
					files.push(filePath);
				}else{
				//	logger.trace("No match file");
				}
			}else if(file.isDirectory && !file.name.match(/^\./)){
				var dirFiles = this.getFilteredFileList([initialRoot, file], regExpFilter, makeUnixPaths, true);
				files.push.apply(files, dirFiles);
			}else{
			//	logger.trace("no match dir");
			}
		}
	}

	return files; //Array
}


fileUtil.copyDir = function(/*String*/srcDir, /*String*/destDir, /*RegExp|String*/regExpFilter, /*boolean?*/onlyCopyNew){
	// logger.trace("copyDir: " + srcDir + " to " + destDir);
	//summary: copies files from srcDir to destDir using the regExpFilter to determine if the
	//file should be copied. Returns a list file name strings of the destinations that were copied.
	srcDir = fileUtil.getFileObj(srcDir);
	destDir = fileUtil.getFileObj(destDir);
	logger.info("Copy directory " + srcDir.nativePath + " to " + destDir.nativePath);
	var fileNames = fileUtil.getFilteredFileList(srcDir, regExpFilter, true);
	var copiedFiles = [];
	
	for(var i = 0; i < fileNames.length; i++){
		var srcFile = new air.File(fileNames[i]);
		var srcFileName = srcFile.nativePath;
		var destFileName = srcFileName.replace(srcDir.nativePath, destDir.nativePath);
		var destFile = new air.File(destFileName);
		

		if(fileUtil.copyFile(srcFile, destFile, onlyCopyNew)){
			copiedFiles.push(destFileName);
		}
	}

	return copiedFiles.length ? copiedFiles : null; //Array or null
}

fileUtil.copyFile = function(/*String*/srcFileName, /*String*/destFileName, /*boolean?*/onlyCopyNew){
	//summary: copies srcFileName to destFileName. If onlyCopyNew is set, it only copies the file if
	//srcFileName is newer than destFileName. Returns a boolean indicating if the copy occurred.
	var destFile = fileUtil.getFileObj(destFileName);
	var srcFile = fileUtil.getFileObj(srcFileName);

	logger.trace("Src filename: " + srcFile.nativePath);
	logger.trace("Dest filename: " + destFile.nativePath);

	//If onlyCopyNew is true, then compare dates and only copy if the src is newer
	//than dest.
	if(onlyCopyNew){
		if( fileUtil.fileExists(destFile) && destFile.modificationDate >= srcFile.modificationDate){
			return false; //Boolean
		}
	}
	// air.trace("fileUtil: copying file:" + srcFile.nativePath + ", exists? " + fileUtil.fileExists(srcFile)); 
	srcFile.copyTo(destFile, true);
	return true; //Boolean
}

fileUtil.readFile = function(/*String*/path, /*String?*/encoding){
	// air.trace("readFile: " + path);
	//summary: reads a file and returns a string
	encoding = encoding || "utf-8";
	var file = fileUtil.getFileObj(path);
//	logger.trace("Going to read: " + file.nativePath);
	var stream = new air.FileStream();
	try
	{
		stream.open(file, air.FileMode.READ);
	} catch (e) {
		logger.error("Unable to open file " + file.nativePath + " for read");
		throw "Unable to open file " + file.nativePath + " for read";
	}
	try
	{
		var data = stream.readMultiByte(stream.bytesAvailable, encoding);
	} catch(e) {
		logger.error("Unable to read file: " + file.nativePath);
		throw "Unable to read file: " + file.nativePath;
	} finally {
		stream.close();
	}
	return data;
}

fileUtil.saveUtf8File = function(/*String*/fileName, /*String*/fileContents){
	//summary: saves a file using UTF-8 encoding.
	fileUtil.saveFile(fileName, fileContents, "utf-8");
}

fileUtil.saveFile = function(/*String*/fileName, /*String*/fileContents, /*String?*/encoding){
	//summary: saves a file.
	logger.trace("Saving " + fileName);
	var outFile = fileUtil.getFileObj(fileName);
	var outWriter;
	if(!encoding){
		encoding = "utf-8";
	}
	outWriter = new air.FileStream();
	outWriter.open(outFile, air.FileMode.WRITE);
	
	outWriter.writeMultiByte(fileContents, encoding);
	outWriter.close();
}

fileUtil.deleteFile = function(/*String*/fileName){
	//summary: deletes a file or directory if it exists.
	var file = fileUtil.getFileObj(fileName);
	if(fileUtil.fileExists(file)){
		if(file.isDirectory){
			file.deleteDirectory(true);
		} else {
			file.deleteFile();
		}
	}
}

fileUtil.mkdir = function(/*String*/ directory){
	var dir = fileUtil.getFileObj(directory);
	dir.createDirectory();
}

// these are the same function in AIR
fileUtil.mkdirs = fileUtil.mkdir;

fileUtil.getParent = function(/*String*/ filename){
	var f = fileUtil.getFileObj(filename);
	return f.parent.nativePath;
}

fileUtil.getAbsolutePath = function(/*String*/ filename){
	//summary: return the absolute path of the file
	var f = fileUtil.getFileObj(filename);
	return f.nativePath;
}

var readText = fileUtil.readFile;